// construct similarity matrix.
#include "Clustering.h"
#include "Methods.h"

MidNode::MidNode(GeneData *data, unsigned int val){
  vector<int> tmp = (*data).get_test();
  vector<int>::iterator it;
  score = 0;
  num = val;
  copy(tmp.begin(), tmp.end(), back_inserter(test_assembly));
  for(it = tmp.begin(); it != tmp.end(); ++it){
    if((*it) == WILD_CARD){ non_zero_count.push_back(0); }
    else{ non_zero_count.push_back(1); }
  }
  newick_format = (*data).get_tracking_id();
  node.push_back(data);
}

MidNode::~MidNode(void){
  vector<int>().swap(test_assembly);
  vector<unsigned int>().swap(non_zero_count);
  list<GeneData *>().swap(node);
}

void MidNode::set_merge_node(list<GeneData *> node1, list<GeneData *> node2){
  node1.splice(node1.end(), node2);
  node.swap(node1);
}

MidNode* MidNode::operator + (const MidNode &obj){
  vector<int> test_this = (*this).get_test_assembly();
  vector<int> test_obj = obj.get_test_assembly();
  vector<unsigned int> count_this = (*this).get_non_zero_count();
  vector<unsigned int> count_obj = obj.get_non_zero_count();
  vector<int>::iterator it_test_this, it_test_obj;
  vector<unsigned int>::iterator it_count_this, it_count_obj;
  MidNode *merge = new MidNode();

  for(it_test_this = test_this.begin(), it_test_obj = test_obj.begin(), it_count_this = count_this.begin(), it_count_obj = count_obj.begin(); it_test_this != test_this.end() && it_test_obj != test_obj.end() && it_count_this != count_this.end() && it_count_obj != count_obj.end(); ++it_test_this, ++it_test_obj, ++it_count_this, ++it_count_obj){
    if((*it_test_this) == (*it_test_obj) && (*it_test_this) != WILD_CARD){ (*merge).push_test_assembly(*it_test_this); }
    else{(*merge).push_test_assembly(WILD_CARD); }
    (*merge).push_non_zero_count((*it_count_this) + (*it_count_obj));
  }

  (*merge).set_score((*this).get_score());
  (*merge).set_newick_format( "(" + (*this).get_newick_format() + ", " +  obj.get_newick_format() + ")" );
  (*merge).set_merge_node((*this).get_node(), obj.get_node());
  (*merge).set_num((*this).get_num(), obj.get_num());

  return merge;
}

MidNode* MidNode::operator + (const Similarity &obj){
  (*this).set_score(1 - obj.get_fraction());
  (*this).add_newick_format((1 - obj.get_fraction()) / 2);
  return this;
}

void MidNode::add_newick_format(double val){
  stringstream str;
  str << ":" << val;
  newick_format += str.str();
}

void MidNode::set_cluster(int val){
  list<GeneData *>::iterator it;
  for(it = node.begin(); it != node.end(); ++it){ (*it)->set_cluster(val); }
}

void MidNode::set_num(unsigned int val1, unsigned int val2){
  if(val1 < val2){ num = val1; }
  else{ num = val2; }
}

Similarity::Similarity(void){
  fraction = -INF;
  exp_dist = INF;
}

Similarity::Similarity(MidNode *node1, MidNode *node2, double val){
  int gain = 0, loss = 0, remainder = 0;
  vector<int> test1 = node1->get_test_assembly();
  vector<int> test2 = node2->get_test_assembly();
  vector<unsigned int> count1 = node1->get_non_zero_count();
  vector<unsigned int> count2 = node2->get_non_zero_count();
  vector<int>::iterator it_test1, it_test2;
  vector<unsigned int>::iterator it_count1, it_count2;
  
  for(it_test1 = test1.begin(), it_test2 = test2.begin(), it_count1 = count1.begin(), it_count2 = count2.begin(); it_test1 != test1.end() && it_test2 != test2.end() && it_count1 != count1.end() && it_count2 != count2.end(); ++it_test1, ++it_test2, ++it_count1, ++it_count2){
    if((*it_test1) == (*it_test2) && (*it_test1) != WILD_CARD){ gain += (*it_count1) + (*it_count2); }
    else{ loss += (*it_count1) + (*it_count2); }
  }
  remainder = gain - loss;
  if(gain == 0 && loss == 0){ fraction = 1; }
  else{ fraction = (double)remainder / ((double)gain + loss); }
  exp_dist = val;
}

bool Similarity::operator < (const Similarity &obj){
  if((*this).get_fraction() < obj.get_fraction()){ return true; }
  else if((*this).get_fraction() == obj.get_fraction() && (*this).get_exp_dist() > obj.get_exp_dist()){ return true; }
  else{ return false; }
}

CalcDistance::CalcDistance(const string distance, const vector<vector<double> > &gene_exp, vector<vector<double> > &exp_dist_table, const unsigned int NUM, const int core){
  if(distance == "pearson"){ calc_pearson(gene_exp, exp_dist_table, NUM, core); }
  else if(distance == "cosine"){ calc_cosine(gene_exp, exp_dist_table, NUM, core); }
  else if(distance == "euclidean"){ calc_euclidean(gene_exp, exp_dist_table, NUM, core); }
  else{ cout << "ERROR: The option about distance is incorrect." << endl; exit(1); }
}

void CalcDistance::calc_pearson(const vector<vector<double> > &gene_exp, vector<vector<double> > &exp_dist_table, const unsigned int NUM, const int core){
  unsigned int i = 0, j = 0;
#pragma omp parallel for num_threads(core) private(j) schedule(guided)
  for(i = 0; i < NUM - 1; ++i){
    for(j = i + 1; j < NUM; ++j){
      exp_dist_table[i][j] = calc_pearson_pair(gene_exp[i], gene_exp[j]);
      exp_dist_table[j][i] = exp_dist_table[i][j];
    }
  }
}

inline double CalcDistance::calc_pearson_pair(vector<double> exp1, vector<double> exp2){
  vector<double>::iterator it_exp1, it_exp2;
  double pearson = 0;
  double covariance = 0;
  double sd1 = 0;
  double sd2 = 0;
  double mean1 = 0;
  double mean2 = 0;
  for(it_exp1 = exp1.begin(); it_exp1 != exp1.end(); ++it_exp1){ mean1 += (*it_exp1); }
  for(it_exp2 = exp2.begin(); it_exp2 != exp2.end(); ++it_exp2){ mean2 += (*it_exp2); }
  mean1 = mean1 / exp1.size();
  mean2 = mean2 / exp2.size();
  
  for(it_exp1 = exp1.begin(), it_exp2 = exp2.begin(); it_exp1 != exp1.end() && it_exp2 != exp2.end(); ++it_exp1, ++it_exp2){
    covariance += ((*it_exp1) - mean1) * ((*it_exp2) - mean2);
    sd1 += pow(((*it_exp1) - mean1), 2);
    sd2 += pow(((*it_exp2) - mean2), 2);
  }
  pearson = 1 - covariance / (sqrt(sd1) * sqrt(sd2));
  return pearson;
}

void CalcDistance::calc_cosine(const vector<vector<double> > &gene_exp, vector<vector<double> > &exp_dist_table, const unsigned int NUM, const int core){
  unsigned int i = 0, j = 0;
#pragma omp parallel for num_threads(core) private(j) schedule(guided)
  for(i = 0; i < NUM - 1; ++i){
    for(j = i + 1; j < NUM; ++j){
      exp_dist_table[i][j] = calc_cosine_pair(gene_exp[i], gene_exp[j]);
      exp_dist_table[j][i] = exp_dist_table[i][j];
    }
  }
}

inline double CalcDistance::calc_cosine_pair(vector<double> exp1, vector<double> exp2){
  double product = 0, s1 = 0, s2 = 0, cosine = 0;
  vector<double>::iterator it_exp1, it_exp2;
  for(it_exp1 = exp1.begin(), it_exp2 = exp2.begin(); it_exp1 != exp1.end() && it_exp2 != exp2.end(); ++it_exp1, ++it_exp2){
    product += (*it_exp1) * (*it_exp2);
    s1 += pow((*it_exp1), 2);
    s2 += pow((*it_exp2), 2);
  }
  cosine = 1 - product / (sqrt(s1) * sqrt(s2));
  return cosine;
}

void CalcDistance::calc_euclidean(const vector<vector<double> > &gene_exp, vector<vector<double> > &exp_dist_table, const unsigned int NUM, const int core){
  unsigned int i = 0, j = 0;
#pragma omp parallel for num_threads(core) private(j) schedule(guided)
  for(i = 0; i < NUM - 1; ++i){
    for(j = i + 1; j < NUM; ++j){
      exp_dist_table[i][j] = calc_euclidean_pair(gene_exp[i], gene_exp[j]);
      exp_dist_table[j][i] = exp_dist_table[i][j];
    }
  }
}

inline double CalcDistance::calc_euclidean_pair(vector<double> exp1, vector<double> exp2){
  double euclidean = 0;
  vector<double>::iterator it_exp1, it_exp2;
  for(it_exp1 = exp1.begin(), it_exp2 = exp2.begin(); it_exp1 != exp1.end() && it_exp2 != exp2.end(); ++it_exp1, ++it_exp2){ euclidean += pow((*it_exp1) - (*it_exp2), 2); }
  return sqrt(euclidean);
}

Clustering::Clustering(vector<MidNode *> *entry, vector<vector<double> > &exp_dist_table, const double t, const bool eval, const int c){
  unsigned int i, j;
  cluster = 0;
  core = c;
  iteration_num = 0;
  threshold = t;
  NUM = entry->size();

  table = new Similarity**[NUM];
  for(i = 0; i < NUM; ++i){
    search_list.push_back(i);
    table[i] = new Similarity*[NUM];
    for(j = 0; j < NUM; ++j){ table[i][j] = NULL; }
  }

  if(eval == true){
    for(i = 0; i < NUM; ++i){ log.insert(make_pair((*entry)[i]->get_newick_format(), vector<int>(1, (*entry)[i]->get_num() + 1))); }
  }

  // fill table.
#pragma omp parallel for num_threads(core) private(j) schedule(guided)
  for(i = 0; i < NUM - 1; ++i){
    for(j = i + 1; j < NUM; ++j){
      table[i][j] = new Similarity((*entry)[i], (*entry)[j], exp_dist_table[i][j]);
    }
  }
}

void Clustering::search_most_similar_nodes(const vector<MidNode *> *entry){
  int i, j, k, l;
  const int SIZE = search_list.size();
  Similarity max;
  Similarity max_vec[SIZE - 1];
  pair<int, int> max_pair[SIZE - 1];

#pragma omp parallel for num_threads(core) private(j, k, l) schedule(guided)
  for(i = 0; i < SIZE - 1; ++i){
    for(j = i + 1; j < SIZE; ++j){
      k = search_list[i];
      l = search_list[j];
      if(max_vec[i] < (*table[k][l])){
	max_vec[i] = (*table[k][l]);
	max_pair[i] = make_pair(k, l);
      }
    }
  }

  for(i = 0; i < SIZE - 1; ++i){
    if(max < max_vec[i]){
      max = max_vec[i];
      first_node = max_pair[i].first;
      second_node = max_pair[i].second;
    }
  }

  first_node_size = (*entry)[first_node]->get_node_size();
  second_node_size = (*entry)[second_node]->get_node_size();
}

void Clustering::merge(vector<MidNode *> *entry){
  MidNode *tmp_node = NULL;
  Similarity *tmp_sim = NULL;
  tmp_sim = table[first_node][second_node];

  // set cluster num.
  if((1 - (*tmp_sim).get_fraction()) > threshold && (*(*entry)[first_node]).get_score() <= threshold){
    ++cluster;
    (*(*entry)[first_node]).set_cluster(cluster);
  }
  if((1 - (*tmp_sim).get_fraction()) > threshold && (*(*entry)[second_node]).get_score() <= threshold){
    ++cluster;
    (*(*entry)[second_node]).set_cluster(cluster);
  }

  // merge mid node.
  tmp_node = (*entry)[first_node];
  (*entry)[first_node] = (*(*entry)[first_node]) + (*tmp_sim);
  (*entry)[second_node] = (*(*entry)[second_node]) + (*tmp_sim);
  (*entry)[first_node] = (*(*entry)[first_node]) + (*(*entry)[second_node]);
  delete tmp_node;

  tmp_node = (*entry)[second_node];
  (*entry)[second_node] = NULL;
  delete tmp_node;
}

//void Clustering::update_second_node(vector<MidNode *> *entry, vector<vector<double> > &exp_dist_table){
void Clustering::update_second_node(vector<vector<double> > &exp_dist_table){
  int i(0);
  Similarity *tmp_sim = NULL;
  vector<int>::iterator it;
  const vector<int>::iterator it_s = find(search_list.begin(), search_list.end(), second_node);
  
  // update similarity table and exp_dist_table.
  for(it = it_s + 1; it != search_list.end(); ++it){
    i = (*it);
    exp_dist_table[second_node][i] = INF;
    exp_dist_table[i][second_node] = INF;
    tmp_sim = table[second_node][i];
    table[second_node][i] = new Similarity();
    delete tmp_sim;
  }
  for(it = search_list.begin(); it != it_s; ++it){
    i = (*it);
    exp_dist_table[i][second_node] = INF;
    exp_dist_table[i][second_node] = INF;
    tmp_sim = table[i][second_node];
    table[i][second_node] = new Similarity();
    delete tmp_sim;
  }
  
  // update search list.
  search_list.erase(find(search_list.begin(), search_list.end(), second_node));
}

void Clustering::logging(vector<MidNode *> *entry){
  int cluster(0);
  string tracking_id;
  list<GeneData *> nodes;
  vector<MidNode *>::iterator it_ent;
  list<GeneData *>::iterator it_n;
  vector<int>::iterator it_s;
  for(it_s = search_list.begin(); it_s != search_list.end(); ++it_s){
    nodes = (*entry)[(*it_s)]->get_node();
    cluster = (*entry)[(*it_s)]->get_num() + 1;
    for(it_n = nodes.begin(); it_n != nodes.end(); ++it_n){
      tracking_id = (*it_n)->get_tracking_id();
      log[tracking_id].push_back(cluster);
    }
  }
}

Preprocessing::Preprocessing(map<string, GeneData *> *gene_list, vector<MidNode *> *entry, vector<vector<double> > &exp_dist_table, const bool DEGs, const string distance, const int core){
  if(DEGs == false){ extract_expressed_genes(gene_list, entry); }
  else{ extract_DEGs(gene_list, entry); }
  make_distance_table(entry, exp_dist_table, distance, core);
}

void Preprocessing::extract_expressed_genes(map<string, GeneData *> *gene_list, vector<MidNode *> *entry){
  unsigned int i(0);
  GeneData *data;
  map<string, GeneData *>::iterator it;

  // entry genes that are expressed at least one condition.
  for(it = gene_list->begin(); it != gene_list->end(); ++it){
    if((*it).second->sign_exp() == true){
      data = it->second;
      it->second = NULL;
      MidNode *mid = new MidNode(data, i);
      entry->push_back(mid);
      gene_exp.push_back(data->get_exp());
      ++i;
    }
  }
  cout << "Num. of expressed genes\t\t : " << entry->size() << endl;
  free_gene_list(gene_list);
}

void Preprocessing::extract_DEGs(map<string, GeneData *> *gene_list, vector<MidNode *> *entry){
  unsigned int i(0);
  GeneData *data;
  map<string, GeneData *>::iterator it;

  // entry DEGs.
  for(it = gene_list->begin(); it != gene_list->end(); ++it){
    if((*it).second->sign_test() == true){
      data = it->second;
      it->second = NULL;

      MidNode *mid = new MidNode(data, i);
      entry->push_back(mid);
      gene_exp.push_back(data->get_exp());
      ++i;
    }
  }
  cout << "Num. of DEGs\t\t\t : " << entry->size() << endl;
  free_gene_list(gene_list);
}

void Preprocessing::make_distance_table(vector<MidNode *> *entry, vector<vector<double> > &exp_dist_table, const string distance, const int core){
  unsigned int i(0);
  const unsigned int NUM = entry->size();
  for(i = 0; i < NUM; ++i){ exp_dist_table.push_back(vector<double>(NUM, INF)); }
  CalcDistance(distance, gene_exp, exp_dist_table, NUM, core);
  vector<vector<double> >().swap(gene_exp);
}

MidNode *cluster_analysis(vector<MidNode *> *entry, vector<vector<double> > &exp_dist_table, map<string, vector<int> > &log, const string method, const double threshold, const bool eval, const int core){
  int cluster(0);
  MidNode *result;

  // clustering.
  if(method == "average"){
    Average *p = new Average(entry, exp_dist_table, threshold, eval, core);
    cluster = p->get_cluster();
    if(eval == true){ log = p->get_log(); }
    delete p;
  }else if(method == "single"){
    Single *p = new Single(entry, exp_dist_table, threshold, eval, core);
    cluster = p->get_cluster();
    if(eval == true){ log = p->get_log(); }
    delete p;
  }else if(method == "complete"){
    Complete *p = new Complete(entry, exp_dist_table, threshold, eval, core);
    cluster = p->get_cluster();
    if(eval == true){ log = p->get_log(); }
    delete p;
  }else if(method == "ward"){
    Ward *p = new Ward(entry, exp_dist_table, threshold, eval, core);
    cluster = p->get_cluster();
    if(eval == true){ log = p->get_log(); }
    delete p;
  }else{
    cout << "ERROR: The option about merge method is incorrect." << endl;
    exit(1);
  }

  if(cluster == 0){
    ++cluster;
    (*(*entry).front()).set_cluster(cluster);
  }
  cout << "Num. of Cluster\t\t\t : " << cluster << endl;

  result = (*entry).front();
  return result;
}

void free_gene_list(map<string, GeneData *> *gene_list){
  map<string, GeneData *>::iterator it;
  for(it = gene_list->begin(); it != gene_list->end(); ++it){ delete (*it).second; }
  delete gene_list;
}
