// clustering.
#ifndef CLUSTERING_H
#define CLUSTERING_H
#include <iostream>
#include <sstream>
#include <string>
#include <map>
#include <vector>
#include <list>
#include <cstdlib>
#include <omp.h>
#include <cmath>
#include <utility>
#include <algorithm>
#include "Input.h"

#define UP 1
#define DOWN -1
#define WILD_CARD 99
#define INF 10000000

using namespace std;

class Similarity;

class MidNode{
 protected:
  unsigned int num;          // assigned for each mid-node.
  double score;              // 1 - (gain - loss) / (gain + loss) in this node.
  string newick_format;
  vector<int> test_assembly;
  vector<unsigned int> non_zero_count;
  list<GeneData *> node;

 public:
  MidNode(void){ num = INF; score = INF; };
  MidNode(GeneData *data, unsigned int val);
  ~MidNode(void);
  MidNode* operator + (const MidNode &obj);
  MidNode* operator + (const Similarity &obj);
  void set_score(double val){ score = val; }
  void set_merge_node(list<GeneData *> node1, list<GeneData *> node2);
  void push_test_assembly(int val){ test_assembly.push_back(val); }
  void push_non_zero_count(unsigned int val){ non_zero_count.push_back(val); }
  void set_newick_format(string str){ newick_format = str; }
  void add_newick_format(double val);
  void set_cluster(int val);
  void set_num(unsigned int val){ num = val; }
  void set_num(unsigned int val1, unsigned int val2);

  double get_score(void) const { return score; }
  vector<int> get_test_assembly(void) const { return test_assembly; }
  vector<unsigned int> get_non_zero_count(void) const { return non_zero_count; }
  string get_newick_format(void) const { return newick_format; }
  unsigned int get_num(void) const { return num; }
  unsigned int get_node_size(void) const { return node.size(); }
  list<GeneData *> get_node(void) const { return node; }
};

class Similarity{
 protected:
  double fraction;         // (gain - loss) / (gain + loss)
  double exp_dist;

 public:
  Similarity(void);
  Similarity(MidNode *node1, MidNode *node2, double val);
  bool operator < (const Similarity &obj);
  void set_fraction(double val){ fraction = val; }
  void set_exp_dist(double val){ exp_dist = val; }
  double get_fraction(void) const { return fraction; }
  double get_exp_dist(void) const { return exp_dist; }
};

class CalcDistance{
 public:
  CalcDistance(void){}
  CalcDistance(const string distance, const vector<vector<double> > &gene_exp, vector<vector<double> > &exp_dist_table, const unsigned int NUM, const int core);
  void calc_pearson(const vector<vector<double> > &gene_exp, vector<vector<double> > &exp_dist_table, const unsigned int NUM, const int core);
  void calc_cosine(const vector<vector<double> > &gene_exp, vector<vector<double> > &exp_dist_table, const unsigned int NUM, const int core);
  void calc_euclidean(const vector<vector<double> > &gene_exp, vector<vector<double> > &exp_dist_table, const unsigned int NUM, const int core);
  inline double calc_pearson_pair(vector<double> exp1, vector<double> exp2);
  inline double calc_cosine_pair(vector<double> exp1, vector<double> exp2);
  inline double calc_euclidean_pair(vector<double> exp1, vector<double> exp2);
};

class Preprocessing{
 protected:
  vector<vector<double> > gene_exp;

 public:
  Preprocessing(void){}
  Preprocessing(map<string, GeneData *> *gene_list, vector<MidNode *> *entry, vector<vector<double> > &exp_dist_table, const bool DEGs, const string distance, const int core);
  ~Preprocessing(void){ vector<vector<double> >().swap(gene_exp); }
  void extract_expressed_genes(map<string, GeneData *> *gene_list, vector<MidNode *> *entry);
  void extract_DEGs(map<string, GeneData *> *gene_list, vector<MidNode *> *entry);
  void make_distance_table(vector<MidNode *> *entry, vector<vector<double> > &exp_dist_table, const string distance, const int core);
};

class Clustering{
 protected:
  int cluster;
  int first_node;                 // 1st node of the most similar nodes.
  int second_node;                // 2nd node of the most similar nodes.
  int core;
  unsigned int first_node_size;   // Num. of genes contained in 1st node.
  unsigned int second_node_size;  // Num. of genes contained in 2nd node.
  unsigned int iteration_num;
  unsigned int NUM;
  double threshold;
  vector<int> search_list;
  Similarity ***table;
  map<string, vector<int> > log;
  
 public:
  Clustering(void){}
  Clustering(vector<MidNode *> *entry, vector<vector<double> > &exp_dist_table, const double t, const bool eval, const int c);
  ~Clustering(void){ delete [] table; }
  void search_most_similar_nodes(const vector<MidNode *> *entry);
  void merge(vector<MidNode *> *entry);
  //void update_second_node(vector<MidNode *> *entry, vector<vector<double> > &exp_dist_table);
  void update_second_node(vector<vector<double> > &exp_dist_table);
  void logging(vector<MidNode *> *entry);
  int get_cluster(void) const { return cluster; }
  map<string, vector<int> > get_log(void) const { return log; }
};

void free_gene_list(map<string, GeneData *> *gene_list);

MidNode *cluster_analysis(vector<MidNode *> *entry, vector<vector<double> > &exp_dist_table, map<string, vector<int> > &log, const string method, const double threshold, const bool eval, const int core);

#endif
